import numpy as np
from matplotlib import pyplot as plt
from scipy import stats


def normal_smoothing(functionValues, step, kernelStd=1):
    """
    Applies smoothing with a normal kernel to given function
    values via a discrete convolution.

    The output corresponds to the same x-values as the input.

    Parameters
    ----------
    functionValues : float[]
        Array with y-values of a function to be smoothed
    step : float
        Difference between the x-values belonging to adjacent
        entries of `data`.
    kernelStd : float
        Standard deviation of the smoothing kernel

    """
    kernelWidth = kernelStd * 5
    if step == 1:
        kernel = stats.norm.pdf(np.arange(2 * kernelWidth + 1), kernelWidth, kernelStd)
        kernel /= kernel.sum()
        result = np.convolve(functionValues, kernel, domains="same")
        return result
    kernel = stats.norm.pdf(np.arange(-kernelWidth, kernelWidth + step / 2, step), 0, kernelStd)
    kernel /= kernel.sum()
    result = np.convolve(functionValues, kernel, mode="same")
    return result

    result[kernelWidth] += result[:kernelWidth].sum()

    return result[kernelWidth:]


def generate_sample(
    model, observableFun, sampleSize=1000, burnIn=0, dt=1, runFun=None, **observableFunArgs
):
    """
    Generates a sample from a model by recording
    an observable in regular time intervals.

    Note: The model is assumed to be ergodic and in a steady state.
    The sample elements are not fully independent in general.

    Parameters
    ----------
    model : pyformind.Model
        Model from which the sample shall be generated
    observableFun : callable
        Function that takes a model instance and returns
        the observables of interest. May return a 1D array
        (results are assumed to be one sample from a
        random vector) or a 2D array (results are assumed
        to be independent samples from a random vector of the
        size of the first row.
    sampleSize : int
        Denotes how often `fun` is to be applied to `model`.
        Note that the size of the sample will be larger, if
        `fun` returns multiple observations.
    burnIn : int
        Denotes how long the model is run before the sample
        is taken
    dt : float
        Time step between two sampling points
    runFun : callable
        Function that takes a model instance and a time
        and runs the model for that time. If not provided
        or `None`, then `model.run()` will be used.
    **observableFunArgs : keyword arguments
        Keyword arguments passed to `fun`

    """
    if runFun is None:
        runFun = lambda _model, _dt: _model.run(_dt)

    if burnIn:
        for _ in range(int(np.ceil(burnIn / dt))):
            runFun(model, dt)

    runFun(model, dt)
    firstSample = observableFun(model, **observableFunArgs)

    if np.isscalar(firstSample):
        chunkSize, dim = 1, 1
        dtype = type(firstSample)
    else:
        firstSample = np.asarray(firstSample)
        if firstSample.ndim == 1:
            chunkSize, dim = 1, firstSample.size
        else:
            chunkSize, dim = firstSample.shape
        dtype = firstSample.dtype

    sample = np.zeros((sampleSize * chunkSize, dim), dtype=dtype)
    sample[:chunkSize] = firstSample

    for i in range(1, sampleSize):
        runFun(model, dt)
        sample[i * chunkSize : (i + 1) * chunkSize] = observableFun(model, **observableFunArgs)
    return sample


def plot_smoothened(
    data,
    bandwidth=None,
    bins=None,
    reflect=True,
    axis=None,
    shift=0,
    onlyHistogram=False,
):
    """
    Plots a histogram and a smoothed curve for a given data set

    Parameters
    ----------
    data : float[]
        Array with the data
    bandwidth : float
        Standard deviatiation of the applied Gaussian kernel
    bins : float[]
        Edges of bins or number of bins used for the histogram
    reflect : bool
        If `True`, reflecting boundary conditions are assumed.
    axis : axis
        Matplotlib axis object used for plotting.
    shift : float
        Shifts the data before plotting
    onlyHistogram : bool
        If `True`, only the histogram (no smoothed curve) will be plotted.

    """
    integral = (bins is not None and type(bins) == np.ndarray and bins.dtype == int) or (
        bins is None and not (data % 1).any()
    )

    if reflect:
        if integral:
            addedData = np.concatenate((data, -data[data != 0]))
        else:
            addedData = np.concatenate((data, -data))
        addedData += shift
    else:
        addedData = data

    data += shift

    if integral:
        bins = np.arange(data.min() - 0.5, data.max() + 0.5)
    elif bins is None:
        bins = min(max(data.size // 50, 20), 500)

    if bandwidth is None:
        if integral:
            bandwidth = 1
        else:
            bandwidth = 0.01

    if axis is None:
        axis = plt.gca()

    values, bins = axis.hist(data, bins=bins, density=True, alpha=0.5)[:2]

    if not onlyHistogram:
        if not integral:
            bins = 1000

        y, smoothingBins = np.histogram(addedData[np.isfinite(addedData)], bins=bins, density=True)[
            :2
        ]

        if reflect and not integral:
            y *= 2

        color = axis.containers[-1].patches[0].get_facecolor()
        smoothened = normal_smoothing(y, smoothingBins[1] - smoothingBins[0], bandwidth)

        smoothingBins = smoothingBins + (smoothingBins[1] - smoothingBins[0]) / 2
        axis.plot(smoothingBins[:-1], smoothened[-y.size :], color=(*color[:-1], 1))

    if reflect:
        axis.set_xlim((shift, None))

    if not onlyHistogram:
        return smoothingBins[:-1], smoothened[-y.size :]
    else:
        return bins[:-1], values


def analyze_distribution(fun, args=[], kwargs={}, n=10, verbose=True):
    """Return mean, standard deviation, and span of repeated evaluations of a function

    Parameters
    ----------
    fun : callable
        funciton to analyze
    args : list, optional
        list of arguments, by default []
    kwargs : dict, optional
        dictionary of keyword arguments, by default {}
    n : int, optional
        number of function evaluations, by default 10
    verbose : int, optional
        if True, print the information; if >= 2, also print the
        result of each function evaluation. By default True

    Returns
    -------
    tuple[float, float, float]
        mean, standard deviation, and span of the function return values
    """
    results = []
    for _ in range(n):
        result = fun(*args, **kwargs)
        results.append(result)
        if verbose >= 2:
            print(result)

    mean = np.mean(results)
    std = np.std(results, ddof=1)
    span = np.max(results) - np.min(results)
    if verbose:
        print("Mean     = {:7.3f}".format(mean))
        print("Std. dev = {:7.3f}".format(std))
        print("Span     = {:7.3f}".format(span))

    return mean, std, span



# distutils: language=c++
#cython: language_level=3, boundscheck=False, wraparound=False, nonecheck=False
'''
Created on 19.04.2021

@author: fischsam
'''


from cython import nogil

import numpy as np

cimport numpy as np
from libc.stdlib cimport malloc, free

#needed to initialize PyArray_API in order to be able to use it
np.import_array()

def _check_dim(np.ndarray arr, dim, name="Array"):
    if not arr.ndim == dim:
        raise ValueError(name + " does not have the right dimension ({})".format(dim))


cdef double** double_arr_to_ptr(np.ndarray[double, ndim=2] array):
    cdef double **result 
    result = <double **> malloc(sizeof(result[0]) * array.shape[0])
    for i in range(array.shape[0]):
        result[i] = <double *> &array[i, 0]
    
    return result

cdef float** float_arr_to_ptr(np.ndarray[float, ndim=2] array):
    cdef float **result 
    result = <float **> malloc(sizeof(result[0]) * array.shape[0])
    for i in range(array.shape[0]):
        result[i] = <float *> &array[i, 0]
    
    return result
    

cdef class ElementaryLikelihoodComputer():
    
    def __init__(self, 
            np.ndarray observations, 
            np.ndarray weights, 
            np.ndarray consideredColumns, 
            np.ndarray domains, 
            np.ndarray inverseBandwidth, 
            double logNormalization,
            double guaranteedLookupDistance,
            biasCorrectionFunction=None,
            dtype=None):
        
        _check_dim(observations, 2, "'observations'")
        _check_dim(weights, 1, "'weights'")
        _check_dim(consideredColumns, 1, "'consideredColumns'")
        _check_dim(domains, 1, "'domains'")
        _check_dim(inverseBandwidth, 1, "'inverseBandwidth'")
        
        rowNumber = observations.shape[0]
        self.columnNumber = columnNumber = observations.shape[1]
        
        if not weights.size == rowNumber:
            raise ValueError("Size of weights does not match "
                             "dimension 2 of observations ({})".format(rowNumber))
        if not inverseBandwidth.size == columnNumber:
            raise ValueError("Size of inverseBandwidth does not match "
                             "dimension 1 of observations ({})".format(columnNumber))
        if not domains.size == columnNumber:
            raise ValueError("Size of domains does not match "
                             "dimension 1 of observations ({})".format(columnNumber))
        if not consideredColumns.size:
            raise ValueError("consideredColumns must contain at least one value.")
        if (consideredColumns >= columnNumber).any() or (consideredColumns < 0).any():
            raise ValueError("consideredColumns must contain nonnegative integers below "
                             "dimension 1 of observations ({})".format(columnNumber))
        if (inverseBandwidth <= 0).any():
            raise ValueError("inverseBandwidth must contain positive values only.")
        if guaranteedLookupDistance <= 0:
            raise ValueError("guaranteedLookupDistance must be positive.")
        
        if dtype: 
            if dtype not in [np.float32, np.float64]:
                raise ValueError("The data type must be either np.float32 or np.float64")
            self.dtype = dtype
        else:
            if observations.dtype == np.float32:
                self.dtype = np.float32
            else:
                self.dtype = np.float64
    
        self.observations = observations.astype(self.dtype, order='C', copy=False)
        self.weights = weights.astype(self.dtype, order='C', copy=False)
        
        self.consideredColumns = consideredColumns.astype(np.int32, order='C', copy=False)
        self.domains = domains.astype(np.int32, order='C', copy=False)
        self.inverseBandwidth = inverseBandwidth.astype(self.dtype, order='C', copy=False)
        self.logNormalization = logNormalization
        self.guaranteedLookupDistance = guaranteedLookupDistance
        self.biasCorrectionFunction = biasCorrectionFunction
        
        cdef:
            int* consideredColumnsData = <int*> self.consideredColumns.data
            int* domainsData = <int*> self.domains.data
            int consideredColumnsSize = self.consideredColumns.size
            int observationsShape = self.observations.shape[0] 
        
        if self.dtype == np.float32:
            self.observationsFloatPointer = float_arr_to_ptr(self.observations)
            with nogil:
                self.grid = construct_grid_float(
                    self.observationsFloatPointer,
                    consideredColumnsData,
                    domainsData,
                    consideredColumnsSize, observationsShape, guaranteedLookupDistance
                    )
        elif self.dtype == np.float64:
            self.observationsDoublePointer = double_arr_to_ptr(self.observations)
            with nogil:
                self.grid = construct_grid_double(
                    self.observationsDoublePointer,
                    consideredColumnsData,
                    domainsData,
                    consideredColumnsSize, observationsShape, guaranteedLookupDistance
                    ) 
        else:
            raise ValueError("Data type not understood. This should not have happened.")
        
    def compute_log_likelihood(self, np.ndarray sample):
        cdef:
            float **sampleFloatPointer
            double **sampleDoublePointer
            np.ndarray out
            
        _check_dim(sample, 2, "'sample'")
        
        if not sample.shape[1] == self.columnNumber:
            raise ValueError("sample does not match observations along axis 1 ({})".format(self.columnNumber))
        
        sample = sample.astype(self.dtype, order='C', copy=False)
        out = np.zeros(self.observations.shape[0], dtype=self.dtype)
        
        cdef: 
            float* floatInverseBandwidthData
            float* floatOutData
            double* doubleInverseBandwidthData
            double* doubleOutData
            
            int *consideredColumnsData = <int*> self.consideredColumns.data
            int *domainsData = <int*> self.domains.data
            int observationsShape = self.observations.shape[0] 
            int sampleShape = sample.shape[0] 
            int consideredColumnsSize = self.consideredColumns.size
        
        if self.dtype == np.float32:
            sampleFloatPointer = float_arr_to_ptr(sample)
            floatInverseBandwidthData = <float*> self.inverseBandwidth.data
            floatOutData = <float*> out.data
            with nogil:
                compute_log_likelihood_float(
                    self.grid[0],
                    self.observationsFloatPointer,
                    sampleFloatPointer,
                    consideredColumnsData,
                    domainsData,
                    floatInverseBandwidthData,
                    self.logNormalization, consideredColumnsSize, 
                    observationsShape, sampleShape, self.guaranteedLookupDistance,
                    floatOutData
                    )
                free(sampleFloatPointer)
        else:
            sampleDoublePointer = double_arr_to_ptr(sample)
            doubleInverseBandwidthData = <double*> self.inverseBandwidth.data
            doubleOutData = <double*> out.data
            
            with nogil:
                compute_log_likelihood_double(
                    self.grid[0],
                    self.observationsDoublePointer,
                    sampleDoublePointer,
                    consideredColumnsData,
                    domainsData,
                    doubleInverseBandwidthData,
                    self.logNormalization, consideredColumnsSize, 
                    observationsShape, sampleShape, self.guaranteedLookupDistance,
                    doubleOutData
                    )
                free(sampleDoublePointer)
        
        if self.biasCorrectionFunction:
            out = self.biasCorrectionFunction(out, sample.shape[0])
            out *= self.weights
        
        return out.sum()
    
    def __dealloc__(self):
        del self.grid
        if self.dtype == np.float32:
            free(self.observationsFloatPointer)
        else:
            free(self.observationsDoublePointer)
        


from .kdelikelihood import (
    get_bandwidth_by_silverman,
    ParallelLikelihoodComputer,
    MultipleDatasetLikelihoodComputer,
    LikelihoodComputer,
)
from .optimize import maximize_log_likelihood
from .analyze import analyze_distribution, plot_smoothened



'''
Created on 19.04.2021

@author: fischsam
'''
from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.unordered_map cimport unordered_map
cimport numpy as np


cdef extern from "_kde_tools_internals.h":
    struct VectorHasher:
        pass

    ctypedef long index_t;
    ctypedef vector[index_t] key_t;
    ctypedef unordered_map[key_t, vector[index_t], VectorHasher] map_t;
    
    map_t *construct_grid_float "construct_grid<float>"(float **observations, int *consideredColumns, int *mode, int dim, long lenObservations, float guaranteedLookupDistance) nogil
    map_t *construct_grid_double "construct_grid<double>"(double **observations, int *consideredColumns, int *mode, int dim, long lenObservations, double guaranteedLookupDistance) nogil
    
    void compute_log_likelihood_float "compute_log_likelihood<float>"(map_t &hashMap, float **observations, float **sample, int *consideredColumns, int *mode, float *inverseBandwidth, float logNormalization, int dim, long lenObservations, long lenSample, float guaranteedLookupDistance, float *out) nogil
    void compute_log_likelihood_double "compute_log_likelihood<double>"(map_t &hashMap, double **observations, double **sample, int *consideredColumns, int *mode, double *inverseBandwidth, double logNormalization, int dim, long lenObservations, long lenSample, double guaranteedLookupDistance, double *out) nogil
    
#     void print_timing()

cdef class ElementaryLikelihoodComputer:
    
    cdef: 
        map_t *grid
        float **observationsFloatPointer
        double **observationsDoublePointer
        np.ndarray observations
        np.ndarray consideredColumns 
        np.ndarray domains
        np.ndarray inverseBandwidth
        np.ndarray weights
        double logNormalization
        double guaranteedLookupDistance
        object dtype
        object biasCorrectionFunction
        long columnNumber


"""
Created on 01.03.2021

@author: fischsam
"""
import os
import threading
from warnings import warn

from scipy import stats
from scipy.integrate import odeint
from scipy.interpolate import interp1d

import matplotlib.pyplot as plt
import numpy as np

from objectproxypool import ProxyPool, SharedArrayWrapper, unpack
from objectproxypool.pool import get_chunksize_per_worker

CPU_COUNT = os.cpu_count()
odeLock = threading.Lock()

from ._kde_tools import ElementaryLikelihoodComputer


def get_bandwidth_by_silverman(data, nanVal=np.nan, dataGroups=1, sampleSize=None):
    """
    Returns kernel density bandwidths based on a Rule of thumb
    by Silverman, B.W.: Density estimation for statistics and data analysis (Eq. 3.28)
    for Gaussialn kernels.

    Parameters
    ----------
    data : float[][]
        Array of multivariate observations
    nanVal : float
        Value for non-existent / not applicable observations
    dataGroups : int
        Number of independent groups that the data will be split into
    sampleSize : float
        Size of the sample for which the bandwidth is sought. If `None`,
        the size of the provided data is used.

    """
    if np.isnan(nanVal):
        data = np.ma.array(data, mask=np.isnan(data))
    elif nanVal:
        data = np.ma.array(data)
        data[data == nanVal] = np.ma.masked

    if sampleSize is None:
        sampleSize = data.shape[0]

    return np.maximum(
        1.06
        * np.asarray(np.std(data, axis=0))
        * (sampleSize / dataGroups) ** (-1 / (4 + data.shape[1])),
        1e-7,
    )


class NoBiasCorrectionFunction:
    """
    Dummy function not correcting any bias.

    Can be used instead of `BiasCorrectionFunction` if nothing
    should be done.
    """

    def __call__(self, x, sampleSize=None):
        return x


class BiasCorrectionFunction:
    """
    Function correcting the bias introduced in the
    kernel density estimate by taking the logarithm.

    The idea is to use a first order approimation for the bias
    and variance. The resulting correction term is derived
    by solving an ODE.
    """

    def __init__(
        self,
        integralsOfSquaredKernels,
        sampleSize=None,
        step=0.001,
        x0=(0, 1),
        minArg=-20,
        steadyStateTolerance=1e-14,
    ):
        """
        Initializer

        Parameters
        ----------
        integralsOfSquaredKernels : float[]
            Array containing the squared integral of the
            kernels used in each dimension. Note that
            the kernel depends on the bandwidth here!
        sampleSize : int
            Size of the sample used to estimate the likelihood.
            This can also be provided later when calling the function.
        step : float
            step size for the solution of the ODE
        x0 : (float, float)
            Initial condition of the ODE. (x(0), x'(0)) This can be changed to
            minimize the mean squared error. The default value is typically good.
            Note that the function becomes discontinueous unless x0[0] = 0 and
            undifferentiable unless x0[1] = 1.
        minArg : float
            Guess for the minimal argument with which this function will be called.
            The guess will be adjusted if necessary.
        steadyStateTolerance : float
            For small arguments, the ODE, which approaches a steady state as x->-inf,
            may not need to be solved. Instead the value of the steady state
            could be used.
        """
        self._prefactor = 2 / np.asarray(integralsOfSquaredKernels).prod()
        self.step = step
        self.x0 = x0
        self.minArg = minArg
        self.sampleSize = sampleSize
        self._hasReachedSteadyState = False
        self.steadyStateTolerance = steadyStateTolerance
        if sampleSize:
            self.update_ode()

    def set_sample_size(self, sampleSize):
        if self.sampleSize == sampleSize:
            return
        self.sampleSize = sampleSize
        self.update_ode()

    def update_ode(self):
        factor = self._prefactor * self.sampleSize

        self.odeFun = lambda xx, tt: (xx[1], (tt - xx[0]) * factor * np.exp(tt) + xx[1])
        self.odeGrad = lambda _, tt: ((0, 1), (-factor * np.exp(tt), 1))
        self.update_solution()

    def update_solution(self):
        t = np.arange(0, self.minArg - self.step / 2, -self.step)

        with odeLock:
            solution = odeint(self.odeFun, self.x0, t, Dfun=self.odeGrad)

        self.solution = interp1d(t[::-1], solution[::-1, 0])
        self._hasReachedSteadyState = np.allclose(
            self.odeFun(solution[-1], self.minArg), 0, atol=self.steadyStateTolerance
        )

    def __call__(self, x, sampleSize=None):
        if sampleSize:
            self.set_sample_size(sampleSize)

        minX = np.min(x)
        for _ in range(20):
            if minX > self.minArg or self._hasReachedSteadyState:
                break
            self.minArg *= 2
            self.update_solution()
        else:
            warn(
                "BiasCorrectionFunction is called for a small argument "
                "but the steady state has not been reached within the given "
                "time. The returned values may be wrong."
            )

        # Since we know that the ODE has an asymptote as x->-infinity,
        # we can approximate small values.
        if minX < self.minArg:
            x = np.maximum(x, self.minArg)

        if np.max(x) > 0:
            if np.isscalar(x):
                return np.log(x)
            result = np.empty_like(x)
            result[x >= 0] = x[x >= 0]
            result[x < 0] = self.solution(x[x < 0])
            return result

        return self.solution(x)


class LikelihoodComputer:
    def __init__(
        self,
        observations,
        weights,
        bandwidths,
        domains=0,
        consideredColumns=None,
        atol=1e-5,
        guaranteedLookupDistance=None,
        transformData=True,
        correctBias=True,
    ):
        if hasattr(domains, "__iter__"):
            domains = np.asarray(domains, dtype=np.int32, order="C")
        elif domains:
            domains = np.full(observations.shape[1], domains, dtype=np.int32, order="C")

        bandwidths = np.asarray(bandwidths)
        if np.isscalar(bandwidths):
            bandwidths = np.full_like(domains, bandwidths, dtype=float)

        if guaranteedLookupDistance is None:
            guaranteedLookupDistance = -np.log(atol)

        self.__guaranteedLookupDistance = guaranteedLookupDistance
        self.__inverseBandwidths = inverseBandwidths = 1 / bandwidths
        self.__transformData = transformData

        if transformData:
            observations = observations * inverseBandwidths

        if consideredColumns is None:
            consideredColumns = np.arange(len(bandwidths))
        elif not len(consideredColumns):
            raise ValueError("consideredColumns must contain at least one value.")

        # sort observations array for potentially improved performance
        order = np.lexsort(observations.T[consideredColumns[::-1]])
        observations = observations[order]
        weights = np.asarray(weights)[order]

        considered = np.zeros_like(observations, dtype=bool)
        considered[:, consideredColumns] = True
        considered &= np.isfinite(observations)

        consideredColumnKeys = list(
            (np.nonzero(row)[0], (considered == row).all(1))
            for row in np.unique(considered, axis=0)
        )

        if not consideredColumnKeys[0][0].size:
            raise ValueError("No data row is allowed to contain NaNs only.")

        kernelWidth = 120
        normalizations = np.zeros_like(bandwidths, dtype=float)
        integralsOfSquaredKernels = np.zeros_like(bandwidths, dtype=float)
        continuous = domains < 2
        kernelVars = bandwidths * bandwidths
        normalizations[continuous] = np.sqrt(2 * np.pi * kernelVars[continuous])
        integralsOfSquaredKernels[continuous] = 1 / (np.sqrt(np.pi) * 2 * bandwidths[continuous])
        normalizations[~continuous] = (
            np.exp(-np.arange(kernelWidth) ** 2 / (2 * kernelVars[~continuous][:, None]))
        ).sum(1) * 2 - 1
        integralsOfSquaredKernels[~continuous] = (
            np.exp(-np.arange(kernelWidth) ** 2 / (kernelVars[~continuous][:, None]))
        ).sum(1) * 2 - 1
        logNormalizations = np.log(normalizations)

        self.__elementaryLikelihoodComputers = {
            tuple(columns): ElementaryLikelihoodComputer(
                observations[rows],
                weights[rows],
                columns,
                domains,
                inverseBandwidths,
                logNormalizations[columns].sum(),
                guaranteedLookupDistance * len(columns) / len(consideredColumns),
                (
                    BiasCorrectionFunction(integralsOfSquaredKernels[columns])
                    if correctBias
                    else NoBiasCorrectionFunction()
                ),
            )
            for columns, rows in consideredColumnKeys
        }

    def compute_log_likelihood(self, sample):
        if self.__transformData:
            sample = sample * self.__inverseBandwidths

        return sum(
            elem.compute_log_likelihood(sample)
            for elem in self.__elementaryLikelihoodComputers.values()
        )


class MultipleDatasetLikelihoodComputer:
    def __init__(
        self,
        observations,
        weights,
        bandwidths,
        modes=0,
        consideredColumns=None,
        atols=1e-3,
        guaranteedLookupDistances=None,
        transformData=True,
        correctBias=True,
    ):
        if consideredColumns is None:
            consideredColumns = [None] * len(observations)
        self.consideredColumns = consideredColumns

        if np.isscalar(bandwidths):
            bandwidths = [bandwidths] * len(observations)
        self.bandwidths = bandwidths

        if np.isscalar(modes):
            modes = [modes] * len(observations)
        self.modes = modes

        if np.isscalar(atols):
            atols = [atols] * len(consideredColumns)
        self.atols = atols

        if np.isscalar(guaranteedLookupDistances) or guaranteedLookupDistances is None:
            guaranteedLookupDistances = [guaranteedLookupDistances] * len(consideredColumns)
        self.guaranteedLookupDistances = guaranteedLookupDistances

        self.likelihoodComputers = [
            LikelihoodComputer(
                obs, we, bw, m, consCol, atol, guaranteedLookupDistance, transformData, correctBias
            )
            for obs, we, bw, m, consCol, atol, guaranteedLookupDistance in zip(
                observations,
                weights,
                bandwidths,
                modes,
                consideredColumns,
                atols,
                guaranteedLookupDistances,
            )
        ]

    def compute_log_likelihood(self, sample):
        return sum(comp.compute_log_likelihood(sample) for comp in self.likelihoodComputers)


class RemoteMultipleDatasetLikelihoodComputer(MultipleDatasetLikelihoodComputer):
    def __init__(self, jobCounter=None):
        pass

    def init(self, args, kwargs={}):
        super().__init__(*args, **kwargs)
        return threading.get_native_id()

    def compute_log_likelihood(self, sample):
        return super().compute_log_likelihood(unpack(sample))


class ParallelLikelihoodComputer:
    """
    An object allowing to compute the KDE likelihood in parallel

    The object holds a set of field data and allows to estimimate
    their likelihood based on a set of model-generated data
    """

    def __init__(
        self,
        observations,
        bandwidths,
        domains=0,
        dataGroups=1,
        correctBias=True,
        atols=1e-3,
        guaranteedLookupDistances=None,
        numWorkers=None,
        separateProcesses=False,
    ):
        """
        Initializer

        Parameters
        ----------
        observations : float[][]
            Array containing the field data for which the likelihood
            should be computed. Each column represents a different data
            feature; each row represents a different observation
        bandwidths : float[]
            Array containing the bandwidth of the kernel density estimation
            for each data feature. The bandwidth should correspond to the
            scale of the data. Larger bandwidths reduce the stochasticity of
            the likelihood estimates; smaller bandwidths reduce the bias of the
            estimates.
        domains : int[]
            Array specifying the domains of the data features. A value of 0
            corresponds to unconstrained real numbers; a value of 1 to
            non-negative real numbers; a value of 2 to natural numbers.
        dataGroups : int
            Specifys into how many independent feature subsets the data set may
            be split. A value of 1 means that all features depend on one another;
            a value as high as the number of features indicates that all
            features are independent. Smaller values lead to correct likelihood
            estimates in the limit of infinite generated data; larger values
            lead to more robust an less bias estimates of the composite likelihood.
        correctBias : bool
            Whether to apply bias correction.
        atols : float[]
            Approximate tolerance (order of magnitude) for the computational error
            in the estimate. Smaller values result in higher accuracy; larger values
            in faster computation time. Can be specified for each data feature
            individually.
        numWorkers : int
            Number of separate threads or processes used. Default is the number of
            available CPUs.
        separateProcesses : weather to use seperate processes (multiprocessing) or
            separate threads (multithreading). The former is often better on systems
            with many (>10) CPUs and if large data sets are considered. The latter
            can be much faster otherwise due to reduced overhead.
        """
        if numWorkers is None:
            numWorkers = os.cpu_count()

        observations = np.asarray(observations)

        if dataGroups > observations.shape[1]:
            raise ValueError("We cannot consider more data groups than data features.")

        self.separateProcesses = separateProcesses

        observationCount, dataDim = observations.shape
        dataSectionLength = int(np.ceil(dataDim / dataGroups))
        dtype = observations.dtype
        if np.issubdtype(dtype, np.integer):
            dtype = np.double

        if np.isscalar(bandwidths):
            bandwidths = np.full(dataDim, bandwidths)
        else:
            bandwidths = np.asarray(bandwidths)

        observations = observations / bandwidths.astype(dtype)

        if np.isscalar(domains):
            domains = np.full(dataDim, domains)
        else:
            domains = np.asarray(domains)

        if np.isscalar(atols):
            atols = [atols] * dataGroups

        if np.isscalar(guaranteedLookupDistances) or guaranteedLookupDistances is None:
            guaranteedLookupDistances = [guaranteedLookupDistances] * dataGroups

        observationListRaw = []

        def argGenerator():
            # prepare data
            nonlocal observationListRaw
            observationListRaw = []
            colListRaw = []
            validObservationsList = []
            countsListRaw = []
            notExistentValue = np.nanmin(observations) - 1
            assert np.isfinite(notExistentValue)
            for i in range(dataGroups):
                # the considered columns
                col = np.arange(i * dataSectionLength, min((i + 1) * dataSectionLength, dataDim))

                # the considered data
                observations_ = observations[:, col].copy()

                # replace nans with some other value
                observations_[np.isnan(observations_)] = notExistentValue

                # remove duplicates (note: data get sorted!)
                _, uniqueIndices, counts = np.unique(
                    observations_, axis=0, return_index=True, return_counts=True
                )

                # shuffle for equal load on the processors
                randomOrder = np.arange(counts.size)
                np.random.shuffle(randomOrder)
                uniqueIndices = uniqueIndices[randomOrder]
                counts = counts[randomOrder]

                # store prepared data
                colListRaw.append(col)
                observationListRaw.append(observations[uniqueIndices].copy())
                validObservationsList.append(~np.isnan(observations[:, col][uniqueIndices]))
                countsListRaw.append(counts)

            # determine workload
            jobNumber = sum(validObservations.sum() for validObservations in validObservationsList)
            taskLengths = get_chunksize_per_worker(jobNumber, numWorkers)
            plannedCumTask = np.cumsum(taskLengths)

            dataIndex = 0  # the current position in the data set
            taskIndex = 0  # the number of already returned elements (tasks)
            scheduledTasks = 0  # the number of already processed rows
            addOne = True
            isNewTask = True

            # TODO: there may be a bug that sometimes causes an empty data set to be included.
            # this happens randomly based on sorting order (at least it looks like it).
            # Restarting helps in this case. A thorough fix is still required, though.

            for (
                col,
                observations_,
                validObservations,
                counts,
                atol,
                guaranteedLookupDistance,
            ) in zip(
                colListRaw,
                observationListRaw,
                validObservationsList,
                countsListRaw,
                atols,
                guaranteedLookupDistances,
            ):
                # process all rows for the current columns
                while dataIndex < observations_.shape[0]:
                    # if we have a fresh task, reset all the information
                    if isNewTask:
                        newlyScheduledTasks = 0
                        wantedTaskLength = taskLengths[taskIndex]
                        taskLength = 0
                        obs = []
                        cts = []
                        atols_ = []
                        cols = []
                        guaranteedLookupDistances_ = []
                        isNewTask = False

                    # look how many data we have to process
                    oldDataIndex = dataIndex
                    taskLengthChart = validObservations[oldDataIndex:].sum(1).cumsum()
                    additionalRowNumber = (
                        np.searchsorted(taskLengthChart, wantedTaskLength - newlyScheduledTasks)
                        + addOne
                    )

                    # we return at least one job
                    additionalRowNumber = max(additionalRowNumber, 1)

                    dataIndex += additionalRowNumber

                    # if we had added one, how many items would we have processed?
                    if additionalRowNumber - addOne < taskLengthChart.size:
                        potentialTaskLength = (
                            taskLength + taskLengthChart[additionalRowNumber - addOne]
                        )
                    else:
                        potentialTaskLength = taskLength + taskLengthChart[-1]

                    obs.append(observations_[oldDataIndex:dataIndex])
                    cts.append(counts[oldDataIndex:dataIndex])
                    cols.append(col)
                    atols_.append(atol)
                    guaranteedLookupDistances_.append(guaranteedLookupDistance)

                    # check how long the job is alreday (processing nans does not count)
                    newTaskCount = np.isfinite(obs[-1][:, col]).sum()
                    scheduledTasks += newTaskCount
                    taskLength += newTaskCount
                    newlyScheduledTasks += newTaskCount

                    # if the job is long enough
                    if (
                        potentialTaskLength >= wantedTaskLength
                        or plannedCumTask[taskIndex] == scheduledTasks
                    ):
                        # if we are not ahead of the schedule, make the task at least as long as wanted
                        addOne = scheduledTasks <= plannedCumTask[taskIndex]

                        # yield result
                        dataSetNumber = len(obs)
                        yield obs, cts, [bandwidths] * dataSetNumber, [
                            domains
                        ] * dataSetNumber, cols, atols_, guaranteedLookupDistances_, False, correctBias

                        # start a new task
                        taskIndex += 1
                        isNewTask = True
                        if dataIndex == observations_.shape[0]:
                            dataIndex = 0
                            break
                    else:
                        dataIndex = 0
                        break

        self.bandwidths = bandwidths.astype(dtype)

        arguments = list(argGenerator())
        self.likelihoodComputers = ProxyPool(
            RemoteMultipleDatasetLikelihoodComputer,
            min(numWorkers, len(arguments)),
            separateProcesses=separateProcesses,
        )
        assert len(arguments) == self.likelihoodComputers.numWorkers
        threadIDs = self.likelihoodComputers.init(
            arguments, synchronize_workers=True, map_args=True
        )
        if not np.unique(threadIDs).size == len(threadIDs):
            raise AssertionError(
                "One of the workers did not get initialized "
                "properly. The results will be wrong. "
                "This may be because one of the threads "
                "got stuck. Just try it again. If this issue occurs "
                "repeatedly, file an issue on the project's bug tracker."
            )

    def compute_log_likelihood(self, sample):
        if self.separateProcesses:
            sample = SharedArrayWrapper(sample)
            sample.array /= self.bandwidths
            return np.sum(
                self.likelihoodComputers.compute_log_likelihood(
                    sample.remoteArray, synchronize_workers=True
                )
            )
        else:
            sample = sample / self.bandwidths
            return np.sum(
                self.likelihoodComputers.compute_log_likelihood(sample, synchronize_workers=True)
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.likelihoodComputers.close()



"""
Created on 25.10.2021

@author: fischsam
"""


from functools import partial
import logging

import pybobyqa
from pybobyqa import controller
from scipy import optimize as op

import numpy as np


ORIGINAL_initialise_coordinate_directions = controller.Controller.initialise_coordinate_directions


class PreconditionedFunction:
    def __init__(
        self,
        fun,
        x0,
        bounds,
        acceptanceBounds=None,
        initStep=0.1,
        maxSteps=5,
        f0=None,
        args=[],
        assertedMinDifference=2.001,
        printStochasticityInfo=False,
        **kwargs
    ):
        self.fun = partial(fun, *args, **kwargs)

        self.scalingFactors = np.ones(len(x0))

        if f0 is None:
            f0 = self(x0)

        if acceptanceBounds is None:
            sample = [fun(x0) for _ in range(10)]
            stdDev = np.std(sample, ddof=1)
            self.stdDev = stdDev
            acceptanceBounds = (2 * stdDev, 10 * stdDev)
            if printStochasticityInfo:
                print("Mean               = {:8.2f}".format(np.mean(sample)))
                print("Standard deviation = {:8.2f}".format(stdDev))
                print("Min                = {:8.2f}".format(np.min(sample)))
                print("Max                = {:8.2f}".format(np.max(sample)))
                print("Range              = {:8.2f}".format(np.max(sample) - np.min(sample)))
        else:
            self.stdDev = None

        minAcceptedDiff, maxAcceptedDiff = acceptanceBounds

        for i, (lowerBound, upperBound) in enumerate(bounds):
            step = (upperBound - lowerBound) * initStep

            currentUpperBound = upperBound - x0[i]
            currentLowerBound = 0

            x = np.array(x0, copy=True)
            reachedBound = False
            for j in range(maxSteps):
                if (
                    np.abs((upperBound - lowerBound) / (currentLowerBound + 1e-200))
                    < assertedMinDifference
                ):
                    break

                if x0[i] + step > upperBound:
                    if x0[i] - step < lowerBound:
                        reachedBound = True
                        if x0[i] + step - upperBound > lowerBound - (x0[i] - step):
                            step = upperBound - x0[i]
                        else:
                            step = x0[i] - lowerBound
                    else:
                        step = -step

                    if step < 0:
                        currentLowerBound = 0
                        currentUpperBound = lowerBound - x0[i]

                x[i] = x0[i] + step

                f = self.fun(x)
                diff = np.abs(f0 - f)
                if diff > maxAcceptedDiff:
                    currentUpperBound = step
                elif diff < minAcceptedDiff:
                    if reachedBound:
                        break
                    currentLowerBound = step
                else:
                    break

                step = (currentUpperBound + currentLowerBound) / 2
            else:
                if diff > maxAcceptedDiff:
                    step = x[i] - x0[i]
                else:
                    step = currentUpperBound

            if np.abs((upperBound - lowerBound) / step) < assertedMinDifference:
                step = (upperBound - lowerBound) / assertedMinDifference

            self.scalingFactors[i] = np.abs(step)

        self.x0 = self.scale_argument(x0)
        self.bounds = self.scale_argument(np.asarray(bounds).T).T

    def scale_argument(self, x):
        """Converts an argument from the original scale to the scale at
        which the preconditioned function works.

        Parameters
        ----------
        x : array[float]
            Argument vector

        Returns
        -------
        array[float]
            Converted argument vector
        """
        return x / self.scalingFactors

    def unscale_argument(self, x):
        """Converts an argument from the scale at which the preconditioned
        function works back to the original scale.

        Parameters
        ----------
        x : array[float]
            Argument vector in the preconditioned scale

        Returns
        -------
        array[float]
            Argument vector in the original scale
        """
        return self.scalingFactors * x

    def __call__(self, x):
        return self.fun(self.unscale_argument(x))


def maximize_log_likelihood(
    logLikelihood,
    bounds,
    x0=None,
    args=[],
    pybobyqaIters=20,
    itersBetweenPreconditioning=5,
    terminationRange=5,
    terminationTolerance=2,
    preconditionedFunctionKwargs={},
    differentialEvolutionKwargs={},
    pybobyqaKwargs={},
    verbose=True,
):
    nfev = 0

    terminationRange = min(terminationRange, pybobyqaIters)

    messages = {
        0: "Optimization terminated successfully",
        1: "Performed the requested number of iterations without satisfying the termination criterion",
        2: "No change between two iterations. Presumably the optimizer got stuck",
    }

    def get_status_str(logL, x):
        return "ln(L) ={:9.2f}, x = [{}]".format(logL, ", ".join("{:6.3f}".format(xx) for xx in x))

    def negativeLogLikelihood(x, *args_, **kwargs_):
        nonlocal nfev
        nfev += 1
        logL = logLikelihood(x, *args, *args_, **kwargs_)
        if verbose >= 2:
            print(get_status_str(logL, x))
        return -logL

    def get_termination_info(results, std):
        logL = np.array([r[0] for r in results[-terminationRange:]])
        resultRange = (logL.min(), logL.max())
        normalizedResultRangeWidth = (resultRange[1] - resultRange[0]) / std
        if len(results) < terminationRange:
            success = False
        else:
            success = normalizedResultRangeWidth < terminationTolerance
        return (success, normalizedResultRangeWidth, resultRange)

    def print_status(heading, logL=None, x=None):
        if not verbose:
            return

        print()
        print("***** {} *****".format(heading))
        if logL is not None:
            print()
            print("ln(L) = ", logL)
            print("x     =", "[{}]".format(", ".join(str(xx) for xx in x)))
        print()

    if x0 is None:
        print_status("Searching a good initial guess")

        differentialEvolutionKwargs_ = dict(
            bounds=bounds, popsize=10, maxiter=10, polish=False, disp=int(verbose)
        )
        differentialEvolutionKwargs_.update(differentialEvolutionKwargs)

        x0Result = op.differential_evolution(
            negativeLogLikelihood,
            **differentialEvolutionKwargs_,
        )

        x0 = x0Result.x

        print_status("Differential evolution result for initial guess", -x0Result.fun, x0)

    preconditionedFunctionKwargs_ = dict(bounds=bounds)
    preconditionedFunctionKwargs_.update(preconditionedFunctionKwargs)
    x = x0

    results = []

    previousX = None
    status = 1
    for iter in range(1, pybobyqaIters + 1):
        if previousX is not None and (np.array(x) == np.array(previousX)).all():
            status = 2
            break

        previousX = x

        if not (iter - 1) % itersBetweenPreconditioning:
            print_status("Preconditioning the likelihood function")
            preconditionedFunc = PreconditionedFunction(
                negativeLogLikelihood, x, **preconditionedFunctionKwargs_
            )

            if verbose:
                print()
                print("Std. dev. of ln(L): {:5.2f}".format(preconditionedFunc.stdDev))

            pybobyqaKwargs_ = dict(
                npt=2 * len(x) + 1,
                maxfun=400,
                seek_global_minimum=True,
                rhobeg=1,
                print_progress=int(verbose),
                user_params={
                    "tr_radius.gamma_dec": 0.9,
                },
            )

            if preconditionedFunc.stdDev:
                pybobyqaKwargs_["user_params"][
                    "noise.additive_noise_level"
                ] = preconditionedFunc.stdDev

        print_status("Starting Py-Bobyqa iteration {}".format(iter))

        pybobyqaKwargs_.update(pybobyqaKwargs)
        iterationResult = pybobyqa.solve(
            preconditionedFunc,
            preconditionedFunc.scale_argument(x),
            bounds=preconditionedFunc.bounds.T,
            **pybobyqaKwargs_,
        )
        x = iterationResult.x = preconditionedFunc.unscale_argument(iterationResult.x)

        results.append((-iterationResult.f, x))

        print_status("Finished Py-Bobyqa iteration {}".format(iter), *results[-1])

        if get_termination_info(results, preconditionedFunc.stdDev)[0]:
            status = 0
            break

    sortedResults = sorted(results)
    fun, x = sortedResults[-1]

    success, normalizedResultRangeWidth, resultRange = get_termination_info(
        results, preconditionedFunc.stdDev
    )

    if verbose:
        print_status("Finished optimization", fun, x)

        print("Message:            ", messages[status])
        print("Success:            ", success)
        print("Status:             {:5d}".format(status))
        print("Iterations:         {:5d}".format(iter))
        print("Likelihood evals:   {:5d}".format(nfev))
        print("Std. dev. of ln(L): {:5.2f}".format(preconditionedFunc.stdDev))
        print(
            "Range of results in the last {} runs: [{:6.2f}, {:6.2f}]".format(
                terminationRange, *resultRange
            )
        )
        print(
            "Width of result range in std. dev.:  {:5.2f} (success threshold: {:4.2f})".format(
                normalizedResultRangeWidth, terminationTolerance
            )
        )
        print()

        print("Results of the individual iterations:")
        print("-------------------------------------")
        for i, (logL, x) in enumerate(results):
            print("iter {:2d}:".format(i), get_status_str(logL, x))

    result = op.OptimizeResult(
        x=x, fun=fun, status=status, success=success, nfev=nfev, nit=iter, message=messages[status]
    )
    return result


def initialise_coordinate_directions_binsearch(
    self,
    number_of_samples,
    num_directions,
    params,
    acceptanceBound,
    maxEvaluationsPerPoint,
    maxValue=np.inf,
):
    if self.do_logging:
        logging.debug("Initialising with coordinate directions")
    # self.model already has x0 evaluated, so only need to initialise the other points
    # num_directions = params("growing.ndirs_initial")
    assert (
        self.model.num_pts <= (self.n() + 1) * (self.n() + 2) // 2
    ), "prelim: must have npt <= (n+1)(n+2)/2"
    assert (
        1 <= num_directions < self.model.num_pts
    ), "Initialisation: must have 1 <= ndirs_initial < npt"

    at_lower_boundary = (
        self.model.sl > -0.01 * self.delta
    )  # sl = xl - x0, should be -ve, actually < -rhobeg
    at_upper_boundary = (
        self.model.su < 0.01 * self.delta
    )  # su = xu - x0, should be +ve, actually > rhobeg

    xpts_added = np.zeros((num_directions + 1, self.n()))
    for k in range(1, num_directions + 1):
        # k = 0 --> base point (xpt = 0)  [ not here]
        # k = 1, ..., 2n --> coordinate directions [1,...,n and n+1,...,2n]
        # k = 2n+1, ..., (n+1)(n+2)/2 --> off-diagonal directions
        if 1 <= k < self.n() + 1:  # first step along coord directions
            dirn = k - 1  # direction to move in (0,...,n-1)
            stepa = self.delta if not at_upper_boundary[dirn] else -self.delta
            stepb = None
            xpts_added[k, dirn] = stepa

        elif self.n() + 1 <= k < 2 * self.n() + 1:  # second step along coord directions
            dirn = k - self.n() - 1  # direction to move in (0,...,n-1)
            stepa = xpts_added[k - self.n(), dirn]
            stepb = -self.delta
            if at_lower_boundary[dirn]:
                stepb = min(2.0 * self.delta, self.model.su[dirn])  # su = xu - x0, should be +ve
            if at_upper_boundary[dirn]:
                stepb = max(-2.0 * self.delta, self.model.sl[dirn])  # sl = xl - x0, should be -ve
            xpts_added[k, dirn] = stepb

        else:  # k = 2n+1, ..., (n+1)(n+2)/2
            # p = (k - 1) % n + 1  # cycles through (1,...,n), starting at 2n+1 --> 1
            # l = (k - 2 * n - 1) / n + 1  # (1,...,1, 2, ..., 2, etc.) where each number appears n times
            # q = (p + l if p + l <= n else p + l - n)
            stepa = None
            stepb = None
            itemp = (k - self.n() - 1) // self.n()
            q = k - itemp * self.n() - self.n()
            p = q + itemp
            if p > self.n():
                p, q = q, p - self.n()  # does swap correctly in Python

            xpts_added[k, p - 1] = xpts_added[p, p - 1]
            xpts_added[k, q - 1] = xpts_added[q, q - 1]

        # Evaluate objective at this new point
        upper = np.zeros_like(xpts_added[k, :])
        upper[xpts_added[k, :] > 0] = self.model.su[xpts_added[k, :] > 0]
        upper[xpts_added[k, :] < 0] = self.model.sl[xpts_added[k, :] < 0]
        lower = np.zeros_like(xpts_added[k, :])

        previous_vals = None
        previous_xpts_added = 0
        previous_diff = 0
        for i in range(maxEvaluationsPerPoint):
            x = self.model.as_absolute_coordinates(xpts_added[k, :])
            f_list, num_samples_run, exit_info = self.evaluate_objective(
                x, number_of_samples, params
            )

            f_mean = np.mean(f_list[:num_samples_run])
            diff = np.abs(f_mean - self.last_run_fopt)

            # if change is bigger than minimal demanded change (noise level)
            if diff >= acceptanceBound[0]:
                if diff <= acceptanceBound[1] or f_mean < self.last_run_fopt:
                    break
                else:
                    if i == maxEvaluationsPerPoint - 1:
                        if previous_vals and f_mean >= maxValue:
                            f_list, num_samples_run, exit_info = previous_vals
                            xpts_added[k, :] = previous_xpts_added
                        break
                    upper = xpts_added[k, :] + 0
            else:
                if i == maxEvaluationsPerPoint - 1 or (upper <= xpts_added[k, :]).all():
                    if previous_vals and previous_diff > diff:
                        f_list, num_samples_run, exit_info = previous_vals
                        xpts_added[k, :] = previous_xpts_added
                    break
                lower = xpts_added[k, :] + 0

            if f_mean < maxValue:
                previous_vals = f_list, num_samples_run, exit_info
                previous_xpts_added = xpts_added[k, :] + 0
                previous_diff = diff

            xpts_added[k, :] = (upper + lower) / 2

        # Handle exit conditions (f < min obj value or maxfun reached)
        if exit_info is not None:
            if num_samples_run > 0:
                self.model.save_point(
                    x,
                    np.mean(f_list[:num_samples_run]),
                    num_samples_run,
                    x_in_abs_coords=True,
                )
            return exit_info  # return & quit

        # Otherwise, add new results (increments model.npt_so_far)
        self.model.change_point(k, x - self.model.xbase, f_list[0])  # expect step, not absolute x
        for i in range(1, num_samples_run):
            self.model.add_new_sample(k, f_extra=f_list[i])

        # If k exceeds N+1, then the positions of the k-th and (k-N)-th interpolation
        # points may be switched, in order that the function value at the first of them
        # contributes to the off-diagonal second derivative terms of the initial quadratic model.
        # Note: this works because the steps for (k) and (k-n) points were in the same coordinate direction
        if self.n() + 1 <= k < 2 * self.n() + 1:
            # Only swap if steps were in different directions AND new pt has lower objective
            if stepa * stepb < 0.0 and self.model.fval(k) < self.model.fval(k - self.n()):
                xpts_added[[k, k - self.n()]] = xpts_added[[k - self.n(), k]]

    return None  # return & continue


def set_pybobyqa_binsearch_init(acceptanceBound, maxEvaluationsPerPoint):
    def newFun(self, number_of_samples, num_directions, params):
        return initialise_coordinate_directions_binsearch(
            self,
            number_of_samples,
            num_directions,
            params,
            acceptanceBound,
            maxEvaluationsPerPoint,
        )

    setattr(controller.Controller, "initialise_coordinate_directions", newFun)


def set_pybobyqa_default_init():
    setattr(
        controller.Controller,
        "initialise_coordinate_directions",
        ORIGINAL_initialise_coordinate_directions,
    )


class use_pybobyqa_binsearch_init(object):
    def __init__(self, acceptanceBound, maxEvaluationsPerPoint):
        set_pybobyqa_binsearch_init(acceptanceBound, maxEvaluationsPerPoint)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        set_pybobyqa_default_init()


if __name__ == "__main__":
    pass



"""
@author: Samuel M. Fischer
"""
from os import path
import os
import re
import subprocess

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext as _build_ext


# ==== Basic settings ====

# Name of the extension
NAME = "_kde_tools"

# Change this to `True` if you want to profile the code
profile = False

# Name of the package (and directory of pyx and pxd files)
packageName = "kdelikelihood"

currentFolder = path.dirname(path.abspath(__file__))

# Read the long description
with open(path.join(currentFolder, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

package_dir = path.join(currentFolder, "src", packageName)

# Change some options, if we want to profile the code
if profile:
    print("Compiling with profiling option switched on!")
    define_macros = [("CYTHON_TRACE", "1")]
    compiler_directives = {"linetrace": True, "binding": True}
    from Cython.Compiler.Options import get_directive_defaults

    directive_defaults = get_directive_defaults()
    directive_defaults["linetrace"] = True
    directive_defaults["binding"] = True
else:
    define_macros = []
    # compiler_directives={'boundscheck':False, 'wraparound':False, "nonecheck":False}
    compiler_directives = {}


def find_vcvarsall():
    # look if vcvarsall is already on the PATH
    try:
        subprocess.call("vcvasall.bat")
        return "vcvasall.bat"
    except FileNotFoundError:
        pass

    results = []
    for programFilesDir in os.environ["PROGRAMFILES"], os.environ["PROGRAMFILES(X86)"]:
        for visualStudioDir in os.scandir(programFilesDir):
            if "Visual Studio" not in visualStudioDir.name:
                continue
            for yearVersionDir in os.scandir(visualStudioDir):
                if not yearVersionDir.name.isnumeric:
                    continue
                for versionDir in os.scandir(yearVersionDir):
                    pathCandidate = path.join(versionDir.path, "VC/Auxiliary/Build/vcvarsall.bat")
                    if path.exists(pathCandidate):
                        results.append(pathCandidate)

    # return the newest version of vcvarsall
    if results:
        return path.normpath(results[-1])

    print(
        "The Visual Studio Compiler file 'vcvarsall.bat' was not found at the place where it was expected.\n"
        "You may find help at https://docs.microsoft.com/de-de/cpp/build/building-on-the-command-line\n"
        "If you do not have Visual Studio installed, please abort with [Ctrl]-[C] and install it."
    )
    while True:
        pathCandidate = input(
            "Please provide the full path to 'vcvarsall.bat' and press [Enter].\n>>> "
        )
        if not pathCandidate.endswith("vcvarsall.bat"):
            pathCandidate = path.join(pathCandidate, "vcvarsall.bat")
        if path.exists(pathCandidate):
            return path.normpath(pathCandidate)
        print(
            "vcvarsall.bat could not be found at the place you specified. "
            "Please try again or press [Ctrl]-[C] to abort."
        )


def remove_dependency_of_python_debug_version(cppExtension):
    # Read in the file
    with open(cppExtension, "r") as file:
        filedata = file.read()

    # Check whether the file has been processed already by searching
    # for a signature string that would occur after processing
    # and is unlikely to be included in unprocesses files
    searchStr = "#undef _DEBUG\s*#include <Python.h>\s*#define _DEBUG"
    if re.search(searchStr, filedata) is not None:
        return

    # Replace the target string
    replaceStr = """
    
        // This section has been manually introduced to avoid dependency
        // on a debug version of python and to fix an issue with VS2022
        // See https://github.com/microsoft/onnxruntime/issues/9735#issuecomment-970718821
        
        #if defined(_MSC_VER)
        #    if (PY_MAJOR_VERSION == 3 && PY_MINOR_VERSION < 4)
        #        define HAVE_ROUND 1
        #    endif
        #    include <corecrt.h>
        #    pragma warning(push)
        #    pragma warning(disable : 4510 4610 4512 4005)
        #endif
        
        #ifdef _DEBUG
          #undef _DEBUG
          #include <Python.h> 
          #define _DEBUG
        #else
          #include <Python.h>
        #endif
        // --------------------------------------------------------------
        
        """
    #    if defined(_DEBUG) && !defined(Py_DEBUG)
    #        define PYBIND11_DEBUG_MARKER
    #        undef _DEBUG
    #    endif
    filedata = re.sub('\s#include "Python.h"\s', replaceStr, filedata)

    # Write the file out again
    with open(cppExtension, "w") as file:
        file.write(filedata)


def insert_vcvars_path(arguments, vcvarsallString="vcvarsall.bat", vcvarsallPath=None):
    for i, item in enumerate(arguments):
        if vcvarsallString in item:
            if vcvarsallPath is None:
                vcvarsallPath = find_vcvarsall()
            arguments[i] = item.replace(vcvarsallString, vcvarsallPath)

    return arguments


# ==== Create a new factory function to import external dependencies AFTER they have been installed ====


class build_ext(_build_ext):
    def finalize_options(self):
        # Call the original function
        _build_ext.finalize_options(self)

        # Import numpy
        # Prevent numpy from thinking it is still in its setup process:
        # __builtins__.__NUMPY_SETUP__ = False
        import numpy

        self.include_dirs.append(numpy.get_include())

        # Cythonize
        from Cython.Build import cythonize

        self.distribution.ext_modules = cythonize(
            self.distribution.ext_modules,
            compiler_directives=compiler_directives,
            language_level="3",
            annotate=True,
            # force=True  # this is required if generated cpp files
            # shall not be reused. This increases
            # compatibility over python version but
            # takes additional time when building
        )


# ==== Collect the cpp source files and folders ====

# List the source directories
include_dirs = [package_dir]


# ==== prepare the Formind compilation ====


# ==== Create extension objects summarizing the information needed for the build ====

extensions = [
    Extension(
        packageName + "." + NAME,
        [path.join("src", packageName, NAME + ".pyx")],
        language="c++",
        include_dirs=include_dirs,
        define_macros=define_macros,  # needed for profiling
    )
]


if __name__ == "__main__":
    setup(
        cmdclass={"build_ext": build_ext},
        ext_modules=extensions,
        packages=[packageName],
        package_data={
            packageName: ["*.pxd", "*.pyx", "*.cpp", "*.c", "*.h"],
        },
    )
